/* MIDI Writing library.  Written by David Madore <david.madore@ens.fr> */

/* Public Domain */

/* $Id: midiwrite.c,v 1.10 2000/07/21 01:40:52 david Exp $ */

#include <stdio.h>
#include <stdlib.h>

#include "midiwrite.h"

#define OOM "Out of memory!\n"

static void
default_fatal (const char *message, void *client_data)
{
  client_data = NULL;
  fprintf (stderr, message);
  exit (1);
}

struct midi_context_s {
  fatal_t fatal;			/* Fatal event handler */
  void *fatal_client_data;		/* Callback for said handler */
  int nbvoices;				/* Number of voices */
  int nbtracks;				/* Number of tracks */
  struct midi_track_s *midi;		/* Midi track data */
  long int gtime;			/* Global time */
  short int subdiv;			/* Subdivision value */
};

struct midi_track_s {
  size_t maxlen;			/* Currently allocated */
  size_t p;				/* Track length */
  unsigned char rs;			/* Running status */
  long int pd;				/* Pending delay */
  long int time;			/* Absolute time */
  unsigned char *d;			/* Track data */
};

#define ALLOC_UNIT 8192			/* One page on many systems */

void
midi_register_fatal (midi_ctx m,
		     fatal_t my_fatal, void *my_client_data)
     /* Register fatal event handler. */
{
  m->fatal = my_fatal;
  m->fatal_client_data = my_client_data;
}

static void
midi_append (midi_ctx m,
	     int trk, unsigned char b)
     /* Append a byte to a MIDI track. */
{
  if ( m->midi[trk].p >= m->midi[trk].maxlen )
    {
      m->midi[trk].maxlen += ALLOC_UNIT;
      m->midi[trk].d = realloc (m->midi[trk].d, m->midi[trk].maxlen);
      if ( ! m->midi[trk].d )
	{
	  m->fatal (OOM, m->fatal_client_data);
	  return;
	}
    }
  m->midi[trk].d[m->midi[trk].p++] = b;
}

static void
midi_append_var (midi_ctx m,
		 int trk, long int v)
     /* Append a variable-length-encoded quantity to a MIDI track. */
{
  int i;

  for ( i=0 ; v>>(7*i) ; i++ ) ;
  if ( i > 0 )
    i--;
  for ( ; i>=0 ; i-- )
    {
      if ( i > 0 )
	midi_append (m, trk, ((v>>(7*i))&0x7f)|0x80);
      else midi_append (m, trk, (v>>(7*i))&0x7f);
    }
}

static void
midi_init_trk (midi_ctx m,
	       int trk)
     /* Initialize and empty a MIDI track. */
{
  m->midi[trk].maxlen = 0;
  m->midi[trk].p = 0;
  m->midi[trk].rs = 0xff;
  m->midi[trk].pd = 0;
  m->midi[trk].time = 0;
  m->midi[trk].d = NULL;
}

midi_ctx
midi_init_with_fatal (int nbvoices_, fatal_t my_fatal, void *my_client_data)
     /* Create and initialize all MIDI tracks. */
{
  int trk;
  midi_ctx m;

  m = malloc (sizeof(struct midi_context_s));
  if ( ! m )
    {
      my_fatal (OOM, my_client_data);
      return NULL;
    }
  m->fatal = my_fatal;
  m->fatal_client_data = my_client_data;
  m->subdiv = 120;
  m->gtime = 0;
  m->nbvoices = nbvoices_;
  m->nbtracks = m->nbvoices + 1;
  m->midi = malloc (m->nbtracks * sizeof(struct midi_track_s));
  if ( ! m->midi )
    {
      m->fatal (OOM, m->fatal_client_data);
      return NULL;
    }
  for ( trk=0 ; trk<m->nbtracks ; trk++ )
    midi_init_trk(m, trk);
  return m;
}

midi_ctx
midi_init (int nbvoices_)
     /* Create and initialize all MIDI tracks (use default fatal event
      * handler). */
{
  return midi_init_with_fatal (nbvoices_, default_fatal, NULL);
}

static void
midi_put_del (midi_ctx m,
	      int trk)
     /* Dump MIDI track's pending delay. */
{
  midi_append_var (m, trk, m->midi[trk].pd);
  m->midi[trk].pd = 0;
}

static void
midi_put_com (midi_ctx m,
	      int trk, unsigned char b)
     /* Append a command to a MIDI track. */
{
  midi_put_del (m, trk);
  if ( ( b >= 0xf0 ) || ( b != m->midi[trk].rs ) )
    {
      midi_append (m, trk, b);
      m->midi[trk].rs = b;
    }
}

void
midi_wait (midi_ctx m,
	   int trk, int ln)
     /* Increase a MIDI track's pending delay. */
{
  m->midi[trk].pd += ln;
  m->midi[trk].time += ln;
  if ( m->midi[trk].time > m->gtime )
    m->gtime = m->midi[trk].time;
}

void
midi_sync (midi_ctx m)
     /* Synchronize all MIDI tracks' notion of time with the furthest
      * one. */
{
  int trk;

  for ( trk=0 ; trk<m->nbtracks ; trk++ )
    midi_wait (m, trk, m->gtime - m->midi[trk].time);
}

void
midi_all_wait (midi_ctx m,
	       int ln)
     /* Increase all MIDI tracks' pending delay - synchronize. */
{
  midi_wait (m, 0, ln);
  midi_sync (m);
}

void
midi_prg (midi_ctx m,
	  int trk, char ch, int prg)
     /* Set program (instrument) on a given channel. */
{
  midi_put_com (m, trk, 0xc0+ch);
  midi_append (m, trk, prg);
}

void
midi_ctl (midi_ctx m,
	  int trk, char ch, char ctl, char val)
     /* Set controller value on a given channel. */
{
  midi_put_com (m, trk, 0xb0+ch);
  midi_append (m, trk, ctl);
  midi_append (m, trk, val);
}

void
midi_n_on (midi_ctx m,
	   int trk, char ch, char nt, char sp)
     /* Turn on note (at speed sp) on a given channel. */
     /* Rk: sp=0 acts as a note off and is generally used instead of
      * it. */
{
  midi_put_com (m, trk, 0x90+ch);
  midi_append (m, trk, nt);
  midi_append (m, trk, sp);
}

void
midi_n_off (midi_ctx m,
	    int trk, char ch, char nt, char sp)
     /* Turn off note (at speed sp) on a given channel. */
{
  midi_put_com (m, trk, 0x80+ch);
  midi_append (m, trk, nt);
  midi_append (m, trk, sp);
}

void
midi_pitchbend (midi_ctx m,
		int trk, char ch, short int val)
     /* Set pitchbend value on a given channel. */
     /* Rk: this may be wrong... */
{
  midi_put_com (m, trk, 0xe0+ch);
  midi_append (m, trk, val & 0x7f);
  midi_append (m, trk, (val>>7)&0x7f);
}

void
midi_set_subdiv (midi_ctx m,
		 short int v)
     /* Set MIDI file global subdivision. */
     /* (The tempo divided by this value determines the MIDI time
      * unit.) */
{
  m->subdiv = v;
}

void
midi_meta_tempo (midi_ctx m,
		 long int v)
     /* Set tempo (on track zero necessarily). */
     /* This may occur at any point but it must occur before any note
      * is played: this is ensured by midi_start(). */
{
  midi_put_com (m, 0, 0xff);
  midi_append (m, 0, 0x51);
  midi_append (m, 0, 3);
  midi_append (m, 0, (v>>16)&0xff);
  midi_append (m, 0, (v>>8)&0xff);
  midi_append (m, 0, v&0xff);
}

void
midi_meta_beat (midi_ctx m,
		char n, char d, char c, char b)
     /* Set beat.  n/2^d is the division indication.  c is the number
      * of MIDI time units per beat.  b is the number of
      * demisemiquavers per beat (? ? ?). */
{
  midi_put_com (m, 0, 0xff);
  midi_append (m, 0, 0x58);
  midi_append (m, 0, 4);
  midi_append (m, 0, n);
  midi_append (m, 0, d);
  midi_append (m, 0, c);
  midi_append (m, 0, b);
}

void
midi_meta_key (midi_ctx m,
	       char s, char mj)
     /* Set key.  s is number of sharps (or negative of number of
      * flats.  mj is 0 if major, 1 if minor. */
{
  midi_put_com (m, 0, 0xff);
  midi_append (m, 0, 0x59);
  midi_append (m, 0, 2);
  midi_append (m, 0, s);
  midi_append (m, 0, mj);
}

static void
midi_meta__text (midi_ctx m,
		 int trk, unsigned char t, size_t len, const char *s)
     /* Include generic text event t on track trk, with data s. */
{
  size_t i;

  midi_put_com (m, trk, 0xff);
  midi_append (m, trk, t);
  midi_append_var (m, trk, len);
  for ( i=0 ; i<len ; i++ )
    midi_append (m, trk, s[i]);
}

void
midi_meta_text (midi_ctx m,
		int trk, size_t len, const char *s)
     /* Text (only at beginning of track) */
{
  midi_meta__text (m, trk, 1, len, s);
}

void
midi_meta_copyright (midi_ctx m,
		     size_t len, const char *s)
     /* Copyright text (only at beginning of track 0) */
{
  midi_meta__text (m, 0, 2, len, s);
}

void
midi_meta_sequence (midi_ctx m,
		    size_t len, const char *s)
     /* Sequence text (only at beginning of track 0) */
{
  midi_meta__text (m, 0, 3, len, s);
}

void
midi_meta_trackname (midi_ctx m,
		     int trk, size_t len, const char *s)
     /* Track name (only at beginning of track) */
{
  midi_meta__text (m, trk, 3, len, s);
}

void
midi_meta_instrument (midi_ctx m,
		      int trk, size_t len, const char *s)
     /* Instrument name (only at beginning of track) */
{
  midi_meta__text (m, trk, 4, len, s);
}

void
midi_meta_lyrics (midi_ctx m,
		  int trk, size_t len, const char *s)
     /* Lyrics text */
{
  midi_meta__text (m, trk, 5, len, s);
}

void
midi_meta_marker (midi_ctx m,
		  int trk, size_t len, const char *s)
     /* Marker text */
{
  midi_meta__text (m, trk, 6, len, s);
}

void
midi_meta_cuepoint (midi_ctx m,
		    int trk, size_t len, const char *s)
     /* Cue point */
{
  midi_meta__text (m, trk, 7, len, s);
}

void
midi_single_note (midi_ctx m,
		  int trk, char ch, char nt, char sp, long int d)
     /* Convenience function: add a single note nt on track trk,
      * channel ch and speed sp for duration d. */
{
  midi_n_on (m, trk, ch, nt, sp);
  midi_wait (m, trk, d);
  midi_n_on (m, trk, ch, nt, 0);
}

void
midi_start (midi_ctx m,
	    long int tempo,
	    char beat_n, char beat_d, char beat_c, char beat_b,
	    char key_s, char key_m)
     /* Start MIDI file: add tempo, beat and key events to the
      * controller track. */
{
  midi_meta_tempo (m, tempo);
  midi_meta_beat (m, beat_n, beat_d, beat_c, beat_b);
  midi_meta_key (m, key_s, key_m);
}

void
midi_finish (midi_ctx m)
     /* Finish MIDI file: add track end event to every track. */
{
  int trk;

  midi_sync (m);
  for ( trk=0 ; trk<m->nbtracks ; trk++)
    {
      midi_put_com (m, trk, 0xff);
      midi_append (m, trk, 0x2f);
      midi_append (m, trk, 0x00);
    }
}

static void
write_word (unsigned short int w, FILE *f)
     /* Write a 16-bit quantity in big-endian format. */
{
  unsigned char b;

  b = (w>>8)&0xff;  fwrite (&b, 1, 1, f);
  b = (w>>0)&0xff;  fwrite (&b, 1, 1, f);
}

static void
write_dword (unsigned long int w, FILE *f)
     /* Write a 32-bit quantity in big-endian format. */
{
  unsigned char b;

  b = (w>>24)&0xff;  fwrite (&b, 1, 1, f);
  b = (w>>16)&0xff;  fwrite (&b, 1, 1, f);
  b = (w>>8)&0xff;  fwrite (&b, 1, 1, f);
  b = (w>>0)&0xff;  fwrite (&b, 1, 1, f);
}

void
midi_dump (midi_ctx m,
	   FILE *f)
     /* Output the MIDI file on the given file (must be open - stdout
      * is fine). */
{
  int trk;

  fprintf (f, "MThd");
  write_dword (6, f);
  write_word (1, f);
  write_word (m->nbtracks, f);
  write_word (m->subdiv, f);
  for ( trk=0 ; trk<m->nbtracks ; trk++)
    {
      fprintf (f, "MTrk");
      write_dword (m->midi[trk].p, f);
      fwrite (m->midi[trk].d, 1, m->midi[trk].p, f);
    }
  fflush (f);
}

void
midi_terminate (midi_ctx m)
     /* Terminate MIDI session.  Free all resources associated with
      * the MIDI data. */
{
  int trk;

  for ( trk=0 ; trk<m->nbtracks ; trk++ )
    free (m->midi[trk].d);
  free (m->midi);
  free (m);
}
