/* MIDI Writing library.  Written by David Madore <david.madore@ens.fr> */

/* Public Domain */

/* $Id: midiwrite.h,v 1.4 2000/07/21 01:40:52 david Exp $ */

typedef struct midi_context_s *midi_ctx;

typedef void (*fatal_t) (const char *message, void *client_data);

void midi_register_fatal (midi_ctx m, fatal_t my_fatal, void *my_client_data);
     /* Register fatal event handler. */
midi_ctx midi_init_with_fatal (int nbvoices_, fatal_t my_fatal, void *my_client_data);
     /* Create and initialize all MIDI tracks. */
midi_ctx midi_init (int nbvoices_);
     /* Create and initialize all MIDI tracks (use default fatal event
      * handler). */
void midi_wait (midi_ctx m, int trk, int ln);
     /* Increase a MIDI track's pending delay. */
void midi_sync (midi_ctx m);
     /* Synchronize all MIDI tracks' notion of time with the furthest
      * one. */
void midi_all_wait (midi_ctx m, int ln);
     /* Increase all MIDI tracks' pending delay - synchronize. */
void midi_prg (midi_ctx m, int trk, char ch, int prg);
     /* Set program (instrument) on a given channel. */
void midi_ctl (midi_ctx m, int trk, char ch, char ctl, char val);
     /* Set controller value on a given channel. */
void midi_n_on (midi_ctx m, int trk, char ch, char nt, char sp);
     /* Turn on note (at speed sp) on a given channel. */
     /* Rk: sp=0 acts as a note off and is generally used instead of
      * it. */
void midi_n_off (midi_ctx m, int trk, char ch, char nt, char sp);
     /* Turn off note (at speed sp) on a given channel. */
void midi_pitchbend (midi_ctx m, int trk, char ch, short int val);
     /* Set pitchbend value on a given channel. */
     /* Rk: this may be wrong... */
void midi_set_subdiv (midi_ctx m, short int v);
     /* Set MIDI file global subdivision. */
     /* (The tempo divided by this value determines the MIDI time
      * unit.) */
void midi_meta_tempo (midi_ctx m, long int v);
     /* Set tempo (on track zero necessarily). */
     /* This may occur at any point but it must occur before any note
      * is played: this is ensured by midi_start(). */
void midi_meta_beat (midi_ctx m, char n, char d, char c, char b);
     /* Set beat.  n/2^d is the division indication.  c is the number
      * of MIDI time units per beat.  b is the number of
      * demisemiquavers per beat (? ? ?). */
void midi_meta_key (midi_ctx m, char s, char mj);
     /* Set key.  s is number of sharps (or negative of number of
      * flats.  mj is 0 if major, 1 if minor. */
void midi_meta_text (midi_ctx m, int trk, size_t len, const char *s);
     /* Text (only at beginning of track) */
void midi_meta_copyright (midi_ctx m, size_t len, const char *s);
     /* Copyright text (only at beginning of track 0) */
void midi_meta_sequence (midi_ctx m, size_t len, const char *s);
     /* Sequence text (only at beginning of track 0) */
void midi_meta_trackname (midi_ctx m, int trk, size_t len, const char *s);
     /* Track name (only at beginning of track) */
void midi_meta_instrument (midi_ctx m, int trk, size_t len, const char *s);
     /* Instrument name (only at beginning of track) */
void midi_meta_lyrics (midi_ctx m, int trk, size_t len, const char *s);
     /* Lyrics text */
void midi_meta_marker (midi_ctx m, int trk, size_t len, const char *s);
     /* Marker text */
void midi_meta_cuepoint (midi_ctx m, int trk, size_t len, const char *s);
     /* Cue point */
void midi_single_note (midi_ctx m, int trk, char ch, char nt, char sp, long int d);
     /* Convenience function: add a single note nt on track trk,
      * channel ch and speed sp for duration d. */
void midi_start (midi_ctx m, long int tempo,
		 char beat_n, char beat_d, char beat_c, char beat_b,
		 char key_s, char key_m);
     /* Start MIDI file: add tempo, beat and key events to the
      * controller track. */
void midi_finish (midi_ctx m);
     /* Finish MIDI file: add track end event to every track. */
void midi_dump (midi_ctx m, FILE *f);
     /* Output the MIDI file on the given file (must be open - stdout
      * is fine). */
void midi_terminate (midi_ctx m);
     /* Terminate MIDI session.  Free all resources associated with
      * the MIDI data. */
